# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import base64
import os

import requests
from kubernetes import client as kubernetes_client

import fluxburst.defaults as defaults
import fluxburst.utils as utils


def get_minicluster(
    command,
    size=None,
    tasks=None,  # nodes * cpu per node, where cpu per node is vCPU / 2
    cpu_limit=None,
    memory_limit=None,
    flags=None,
    name=None,
    namespace=None,
    image=None,
    wrap=None,
    log_level=7,
    flux_user=None,
    lead_host=None,
    lead_port=None,
    broker_toml=None,
    munge_secret_name=None,
    curve_cert_secret_name=None,
    curve_cert=None,
    lead_size=None,
    lead_jobname=None,
    zeromq=False,
    quiet=False,
    strict=False,
):
    """
    Get a MiniCluster CRD as a dictionary

    Limits should be slightly below actual pod resources. The curve cert and broker config
    are required, since we need this external cluster to connect to ours!
    """
    flags = flags or "-ompi=openmpi@5 -c 1 -o cpu-affinity=per-task"
    image = image or "ghcr.io/flux-framework/flux-restful-api"
    container = {"image": image, "command": command, "resources": {}}

    # Are we bursting?
    doing_burst = lead_port and lead_size and lead_jobname
    if cpu_limit is None and memory_limit is None:
        del container["resources"]
    elif cpu_limit is not None or memory_limit is not None:
        container["resources"] = {"limits": {}, "requests": {}}
    if cpu_limit is not None:
        container["resources"]["limits"]["cpu"] = cpu_limit
        container["resources"]["requests"]["cpu"] = cpu_limit
    if memory_limit is not None:
        container["resources"]["limits"]["memory"] = memory_limit
        container["resources"]["requests"]["memory"] = memory_limit

    # Do we have a custom flux user for the container?
    if flux_user:
        container["flux_user"] = {"name": flux_user}

    # The MiniCluster has the added name and namespace
    mc = {
        "size": size,
        "namespace": namespace,
        "name": name,
        "interactive": False,
        "logging": {"zeromq": zeromq, "quiet": quiet, "strict": strict},
        "flux": {
            "option_flags": flags,
            "connect_timeout": "5s",
            "log_level": log_level,
        },
    }

    # Are we bursting?
    # Providing the lead broker and port points back to the parent
    if doing_burst:
        mc["flux"]["bursting"] = {
            "lead_broker": {
                "address": lead_host,
                "port": int(lead_port),
                "name": lead_jobname,
                "size": int(lead_size),
            },
            "clusters": [{"size": size, "name": name}],
        }

    if tasks is not None:
        mc["tasks"] = tasks

    # This is text directly in config
    if doing_burst and curve_cert:
        mc["flux"]["curve_cert"] = curve_cert

    # A provided secret will take precedence
    if doing_burst and curve_cert_secret_name:
        mc["flux"]["curve_cert_secret"] = curve_cert_secret_name

    # This is just the secret name
    if doing_burst and munge_secret_name:
        mc["flux"]["munge_secret"] = munge_secret_name
    if broker_toml:
        mc["flux"]["broker_config"] = broker_toml

    # eg., this would require strace "strace,-e,network,-tt"
    if wrap is not None:
        mc["flux"]["wrap"] = wrap
    return mc, container


def ensure_curve_cert(curve_cert):
    """
    Ensure we are provided with an existing curve certificate we can load.
    """
    if not curve_cert or not os.path.exists(curve_cert):
        raise ValueError(
            f"Curve cert (provided as {curve_cert}) needs to be defined and exist."
        )
    return utils.read_file(curve_cert)


def ensure_flux_operator_yaml(flux_operator_yaml):
    """
    Ensure we are provided with the installation yaml and it exists!
    """
    # flux operator yaml default is current from main
    if not flux_operator_yaml:
        flux_operator_yaml = utils.get_tmpfile(prefix="flux-operator") + ".yaml"
        r = requests.get(defaults.flux_operator_yaml, allow_redirects=True)
        utils.write_file(r.text, flux_operator_yaml)

    # Ensure it really really exists
    flux_operator_yaml = os.path.abspath(flux_operator_yaml)
    if not os.path.exists(flux_operator_yaml):
        raise ValueError(f"{flux_operator_yaml} does not exist.")
    return flux_operator_yaml


def create_secret(path, secret_path, name, namespace, mode="r"):
    """
    Create a secret
    """
    # Configureate ConfigMap metadata
    metadata = kubernetes_client.V1ObjectMeta(
        name=name,
        namespace=namespace,
    )
    # Get File Content
    with open(path, mode) as f:
        content = f.read()

    # base64 encoded string
    if mode == "rb":
        content = base64.b64encode(content).decode("utf-8")

    # Instantiate the configmap object
    return kubernetes_client.V1Secret(
        api_version="v1",
        kind="Secret",
        type="opaque",
        string_data={secret_path: content},
        metadata=metadata,
    )



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


def is_burstable(jobinfo):
    """
    Determine if a job is burstable if:

    1. It is flagged as burstable
    """
    return "burstable" in jobinfo["spec"]["attributes"]["system"]



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


def in_order(plugins):
    """
    Plugins is expected to be an ordered dict, return in order

    This is designed intending to be flexible. We can add
    additional metadata if needed
    """
    for name, plugin in plugins.items():
        yield name, plugin



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


import os
import socket

from fluxoperator.client import FluxMiniCluster
from kubernetes import client as kubernetes_client
from kubernetes import utils as k8sutils
from kubernetes.client.rest import ApiException

import fluxburst.kubernetes.cluster as helpers
from fluxburst.logger import logger
from fluxburst.plugins import BurstPlugin


class KubernetesBurstPlugin(BurstPlugin):
    """
    An additional wrapper to the plugin that adds support for the Flux Operator
    """

    def ensure_namespace(self, kubectl):
        """
        Use the instantiated kubectl to ensure the cluster namespace exists.
        """
        try:
            kubectl.create_namespace(
                kubernetes_client.V1Namespace(
                    metadata=kubernetes_client.V1ObjectMeta(name=self.params.namespace)
                )
            )
        except Exception:
            logger.warning(
                f"🥵️ Issue creating namespace {self.params.namespace}, assuming already exists."
            )

    def validate(self):
        """
        Validate ensures that required conditions are met.

        This should return True/False to indicate if the plugin is valid or not.
        """
        # If we aren't doing an isolated burst, we require host names, configs, etc.
        if not self.params.isolated_burst:
            # Lead host and port are required. A custom broker.toml can be provided,
            # but we are having the operator create it for us
            if (
                not self.params.lead_port
                or not self.params.lead_host
                or not self.params.lead_size
            ):
                logger.warning(
                    "Parameteres lead_host, lead_size, and lead_port must be defined when not doing an isolated_burst."
                )
                return False

            print(
                f"Broker lead will be accessible on {self.params.lead_host}:{self.params.lead_port}"
            )

            # We also need to have the munge key and curve cert
            if self.params.munge_key and not os.path.exists(self.params.munge_key):
                logger.warning(
                    f"Provided munge key {self.params.munge_key} does not exist."
                )
                return False
        return True

    def check_configs(self):
        """
        Ensure we have required config files.
        """
        message = "does not exist or you don't have permissions to see it"
        if not os.path.exists(self.params.munge_key):
            raise ValueError(f"Provided munge key {self.params.munge_key} {message}.")

        if not os.path.exists(self.params.curve_cert):
            raise ValueError(f"Provided curve cert {self.params.munge_key} {message}.")

    def install_flux_operator(self, kubectl, flux_operator_yaml):
        """
        Install the flux operator yaml
        """
        try:
            k8sutils.create_from_yaml(kubectl.api_client, flux_operator_yaml)
            logger.info("Installed the operator.")
        except Exception as exc:
            logger.warning(
                f"Issue installing the operator: {exc}, assuming already exists"
            )

    def run(self, request_burst=False, nodes=None, tasks=None):
        """
        Given some set of scheduled jobs, run bursting.

        If request_burst is True, a specific number of tasks and size
        must be provided to request a cluster in advance (that jobs can
        be run on).
        """
        # Exit early if no jobs to burst
        if not self.jobs and not request_burst:
            logger.info(f"Plugin {self.name} has no jobs to burst.")
            return

        # If we have requested a burst, tasks and size are required
        if request_burst and (not tasks or not nodes):
            logger.warning("Burst requests require nodes and tasks.")
            return

        # Ensure we have a flux operator yaml file, fosho, foyaml!
        foyaml = helpers.ensure_flux_operator_yaml(self.params.flux_operator_yaml)

        # lead host / port / size / are required in the dataclass
        # But ONLY if this isnt' an isolated burst!
        # We check munge paths here, because could be permissions issue
        if not self.params.isolated_burst:
            self.check_configs()

        # This call registers the name of the cluster to self.clusters
        cli = self.create_cluster()
        kubectl = cli.get_k8s_client()

        # Install the operator!
        self.install_flux_operator(kubectl, foyaml)

        # Are we requesting or running jobs?
        if request_burst:
            self.create_minicluster(kubectl, "sleep infinity", nodes, tasks)
        else:
            self.run_jobs(kubectl)

    def run_jobs(self, kubectl):
        """
        Run jobs (creating MiniClusters), assuming that the burst has been done.
        """
        # Create a MiniCluster for each job
        for _, job in self.jobs.items():
            command = " ".join(job["spec"]["tasks"][0]["command"])
            self.create_minicluster(kubectl, command, job["nnodes"], job["ntasks"])

    def create_minicluster(self, kubectl, command, nodes, tasks):
        """
        Create the MiniCluster
        """
        logger.info(f"Preparing MiniCluster for {command}")

        # The plugin is assumed to be running from the lead broker
        # of the cluster it is bursting from, this we get info about it
        podname = socket.gethostname()
        hostname = podname.rsplit("-", 1)[0]

        # TODO: we are using defaults for now, but will update this to be likely
        # configured based on the algorithm that chooses the best spec
        minicluster, container = helpers.get_minicluster(
            command,
            name=self.params.name,
            memory_limit=self.params.memory_limit,
            cpu_limit=self.params.cpu_limit,
            namespace=self.params.namespace,
            broker_toml=self.params.broker_toml,
            tasks=tasks,
            size=nodes,
            image=self.params.image,
            wrap=self.params.wrap,
            log_level=self.params.log_level,
            flux_user=self.params.flux_user,
            lead_host=self.params.lead_host,
            lead_port=self.params.lead_port,
            munge_secret_name=self.params.munge_secret_name,
            curve_cert_secret_name=self.params.curve_cert_secret_name,
            lead_jobname=hostname,
            lead_size=self.params.lead_size,
        )
        # Create the namespace
        self.ensure_namespace(kubectl)

        # Let's assume there could be bugs applying this differently
        crd_api = kubernetes_client.CustomObjectsApi(kubectl.api_client)

        # These are not needed if we are doing an isolated burst
        if not self.params.isolated_burst:
            self.ensure_secrets(kubectl)

        # Create the MiniCluster! This also waits for it to be ready
        print(
            f"⭐️ Creating the minicluster {self.params.name} in {self.params.namespace}..."
        )

        # Make sure we provide the core_v1_api we've created
        operator = FluxMiniCluster(core_v1_api=kubectl)
        try:
            return operator.create(**minicluster, container=container, crd_api=crd_api)
        except Exception as e:
            print(f"Issue creating cluster, does it already exist?: {e}")
            return True

    def ensure_secrets(self, kubectl):
        """
        Ensure secrets (munge.key and curve.cert) are ready for a job
        """
        secrets = []

        # kubectl create secret --namespace flux-operator munge-key --from-file=/etc/munge/munge.key
        if self.params.curve_cert:
            secrets.append(
                helpers.create_secret(
                    self.params.curve_cert,
                    "curve.cert",
                    self.params.curve_cert_secret_name,
                    self.params.namespace,
                )
            )

        if self.params.munge_key:
            secrets.append(
                helpers.create_secret(
                    self.params.munge_key,
                    "munge.key",
                    self.params.munge_secret_name,
                    self.params.namespace,
                    mode="rb",
                )
            )

        for secret in secrets:
            try:
                logger.debug(f"Creating secret {secret.metadata.name}")
                kubectl.create_namespaced_secret(
                    namespace=self.params.namespace,
                    body=secret,
                )
            except ApiException as e:
                print(
                    "Exception when calling CoreV1Api->create_namespaced_config_map: %s\n"
                    % e
                )



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import importlib
import os
import pkgutil
from dataclasses import dataclass

import fluxburst.defaults as defaults
from fluxburst.logger import logger

# Executor plugins are externally installed plugins named "snakemake_executor_<name>"
# They should follow the same convention if on pip, snakemake-executor-<name>
burstable_plugins = {
    name.replace(defaults.plugin_prefix, ""): importlib.import_module(name)
    for _, name, _ in pkgutil.iter_modules()
    if name.startswith(defaults.plugin_prefix)
}


@dataclass
class BurstParameters:
    pass


class BurstPlugin:
    """
    The base class for a burst plugin defines needed functions.
    """

    # Default dataclass is essentially empty
    _param_dataclass = BurstParameters

    def __init__(self, dataclass, **kwargs):
        self.set_params(dataclass)

        # Set of jobs assigned to be bursted, and bursted clusters
        self.jobs = {}
        self.clusters = {}

    def schedule(self, job):
        """
        Attempt to schedule a job, if possible.
        """
        raise NotImplementedError

    def run(self, *args, **kwargs):
        """
        Main function to run a plugin with the set of burstable jobs.
        """
        raise NotImplementedError

    def cleanup(self, name=None):
        pass

    def refresh_clusters(self, clusters):
        """
        Update known clusters from a list of those removed.
        """
        updated = {}
        for name in self.clusters:
            if name not in clusters:
                updated[name] = self.clusters[name]
        self.clusters = updated

    def set_params(self, dc):
        """
        Given known parameters, assert we have the correct dataclass
        and update from the environment, etc.

        The dc (dataclass) is used by the plugin as a generic strategy
        to select and move around custom arguments, if needed.
        """
        # The dataclass expected for the plugin must match what is provided
        if not isinstance(dc, self._param_dataclass):
            raise ValueError(
                f"Incorrect dataclass provided, found {dc} and want {self._param_dataclass}"
            )

        # Get params from the environment, which take precedence
        for key, value in os.environ.items():
            if not key.startswith("FLUXBURST_"):
                continue
            key = key.replace("FLUXBURST_", "").lower()
            if hasattr(dc, key):
                logger.debug(f"Found {key} in environment.")
                setattr(dc, key, value)

        # At this point we want to convert the args <dataclasses> back into dataclass
        self.params = dc



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import collections

import fluxburst.handles as handles
import fluxburst.selectors as selectors
import fluxburst.sorting as sorting
from fluxburst.logger import logger

from .plugins import burstable_plugins


class FluxBurst:
    """
    Flux Burst Client
    """

    def __init__(self, handle=None, mock=False, validate=True):
        """
        Create a new burst client.

        Selectors: Process the current queue and add flags/filter jobs for plugins.
        Plugins: take jobs that need bursting, and burst some subset.
        A filter can (on the simplest terms) just look for a job flagged as
        burstable. For more complex cases, it can perform it's own logic and
        flag some subset as burstable instead. Validation is always done for
        the top level module, howevert the validate boolean here determines
        if the plugin should run its own validation function.
        """
        self.reset_selector()
        self.reset_plugins()
        self.flux = handles.FluxMock(handle) if mock else handles.FluxHandle(handle)
        self.set_ordering(sorting.in_order)
        self.validate = validate

    @property
    def choices(self):
        return "|".join(list(self.plugins))

    def load(self, name, dataclass, **kwargs):
        """
        Register a bursting plugin manually.

        We assume a setup does not always want all plugins available.
        The dataclass argument should be for the BurstParameters
        specific to the plugin.
        """
        if name not in burstable_plugins:
            raise ValueError(f"Plugin {name} is not known. Choices are {self.choices}")

        # Validate the plugin, first plugin module then loaded class
        self.validate_module(name)
        plugin = burstable_plugins[name].init(dataclass)

        # We set the name attribute so it's always matched to the module
        plugin.name = name
        self.check_plugin_integrity(plugin)

        # If the plugin does it's own validation, do here
        if not self.validate_plugin(plugin):
            logger.warning(f"Plugin {name} did not validate and cannot be added.")
            return

        self.plugins[name] = plugin

    def validate_module(self, name):
        """
        Validate the structure of the module.

        We currently just require the init function to import and return the class
        """
        if not hasattr(burstable_plugins[name], "init"):
            raise ValueError(
                f"Plugin {name} is missing required init attribute to return BurstPlugin class."
            )

    def validate_plugin(self, plugin):
        """
        Validate a plugin, and if valid, load into module.
        """
        # This should return a boolean for valid or not
        if hasattr(plugin, "validate") and self.validate:
            return plugin.validate()
        return True

    def check_plugin_integrity(self, plugin):
        """
        Validate a plugin, and if valid, load into module.
        """
        required_attributes = ["schedule", "run", "_param_dataclass", "set_params"]
        for required in required_attributes:
            if not hasattr(plugin, required):
                raise ValueError(
                    f"Plugin {plugin.name} is not valid, missing attribute or function {required}"
                )

    def set_selector(self, func):
        """
        Register a job filter or selector for select_jobs.
        """
        self._job_selector = func

    def reset_plugins(self):
        self.plugins = collections.OrderedDict()

    def reset_selector(self):
        self._job_selector = selectors.is_burstable

    def run_burst(self, request_burst=False, nodes=None, tasks=None):
        """
        A full run burst includes:

        1. Selecting jobs using the client filter(s)
        2. In the order of plugins desired, schedule jobs
        3. Return back lookup of scheduled (by plugin)

        If request_burst with nodes and tasks are required, each plugin
        will simply request creation for the size/tasks needed.
        """
        # TODO what to do with unmatched jobs?
        unmatched = self.process_queue()

        # When we get here, run the bursts
        for _, plugin in self.iter_plugins():
            plugin.run(request_burst=request_burst, nodes=nodes, tasks=tasks)
        return unmatched

    def request_burst(self, name, nodes, tasks):
        """
        Request burst is a direct handle to get a plugin and request a burst.

        In this case, we do not create MiniClusters to launch jobs to, but instead
        create a connected cluster anticipating receiving jobs.
        """
        plugin = self.plugins.get(name)
        if not plugin:
            logger.warning(f"Plugin {plugin} is not known.")
            return
        return plugin.run(request_burst=True, nodes=nodes, tasks=tasks)

    def set_ordering(self, func):
        """
        Set a custom function that knows how to yield names, plugins.
        """
        self._iter_func = func

    def iter_plugins(self):
        """
        Custom function to iterate over plugins
        """
        if not self._iter_func:
            self._iter_func = sorting.in_order

        for name, plugin in self._iter_func(self.plugins):
            yield name, plugin

    def process_queue(self):
        """
        Process queue is the entrypoint to run one burst cycle.

        We first apply any selectors to the queue, and then hand the
        resulting filtered jobs off to burstable plugins. There could be
        more fine-tuned logic for directing jobs to plugins.
        """
        # Run filters across queue to select jobs
        # This is a dict, keys with job id, values jobinfo
        jobs = self.select_jobs()

        # Going through plugins, determine if matches and can run
        unmatched = []
        for _, job in jobs.items():
            scheduled = False
            for _, plugin in self.iter_plugins():
                # Give to first burstable plugin that can accept
                if plugin.schedule(job):
                    # Remove the burstable attribute so it isn't assigned to another
                    # This is more for development - we could likely use a better way
                    self.mark_as_scheduled(job, plugin.name)
                    scheduled = True
                    break

            # But if we cannot match, return to caller
            if not scheduled:
                unmatched.append(job)

        if unmatched:
            logger.warning(f"There are {len(unmatched)} jobs that cannot be bursted.")
        return unmatched

    def mark_as_scheduled(self, job, plugin_name):
        """
        Mark a job as scheduled.

        For now, we just remove the burstable attribute to indicate
        that it's been scheduled. We also add the plugin it was scheduled
        with, in case we somehow need to regenerate/remember a burst.
        """
        # Add an attribute that says "this job is assigned to burstable plugin X"
        job["spec"]["attributes"]["system"]["burst-scheduled"] = plugin_name

        # The job is no longer marked as burstable for other plugins
        if "burstable" in job["spec"]["attributes"]["system"]:
            del job["spec"]["attributes"]["system"]["burstable"]

        # Update the KVS (is this possible)?
        # This doesn't currently work, so not doing anything :)
        self.flux.update_jobspec(job)

    def select_jobs(self):
        """
        Use filters to select jobs.

        This step is agnostic to the burstable plugins - it is simply
        deducing (via our selector function) what jobs are contenders
        for bursting. See the README / documentation for example info.
        """
        # Keep track of selected burstable jobs by id
        listing = self.flux.list_jobs()
        selected = {}

        for job in listing.get("jobs", []):
            info = self.flux.get_job_info(job["id"])
            if not self._job_selector(info):
                continue
            print(f"🧋️  Job {job['id']} is marked for bursting.")
            selected[job["id"]] = info

        # We don't give a warning here, because likely there aren't
        # jobs that are burstable (it's a more rare event)
        return selected

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[flux-burst-client]"



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import os
from subprocess import PIPE, STDOUT, Popen

from fluxburst.logger import logger


def ensure_no_extra(extra):
    """
    Ensure no extra arguments (in case typos)
    """
    if extra:
        logger.exit(
            "Extra arguments provided that are not known to this command: %s"
            % " ".join(extra)
        )


def check_install(software, quiet=True, command="--version"):
    """
    check_install will attempt to run the singularity command, and
    return True if installed. The command line utils will not run
    without this check.

    Parameters
    ==========
    software: the software to check if installed
    quiet: should we be quiet? (default True)
    command: the command to use to check (defaults to --version)
    """
    cmd = [software, command]
    try:
        version = run_command(cmd, software)
    except Exception:
        return False
    if version:
        if not quiet and version["return_code"] == 0:
            version = version["message"]
            logger.info("Found %s version %s" % (software.upper(), version))
        return True
    return False


def get_installdir():
    """get_installdir returns the installation directory of the application"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def run_command(cmd, sudo=False, stream=False):
    """run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    """
    stdout = PIPE if not stream else None
    if sudo is True:
        cmd = ["sudo"] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=stdout)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    t = output.communicate()[0], output.returncode
    output = {"message": t[0], "return_code": t[1]}

    if isinstance(output["message"], bytes):
        output["message"] = output["message"].decode("utf-8")

    return output


def confirm_action(question, force=False):
    """confirm if the user wants to perform a certain action

    Parameters
    ==========
    question: the question that will be asked
    force: if the user wants to skip the prompt
    """
    if force is True:
        return True

    response = input(question + " (yes/no)? ")
    while len(response) < 1 or response[0].lower().strip() not in "ynyesno":
        response = input("Please answer yes or no: ")

    if response[0].lower().strip() in "no":
        return False

    return True


def confirm_uninstall(filename, force=False):
    """confirm if the user wants to uninstall a module

    Parameters
    ==========
    filename: the file that will be removed
    force: if the user wants to skip the prompt
    """
    return confirm_action(
        "Are you sure you want to uninstall {}".format(filename), force
    )



from .fileio import (
    can_be_deleted,
    copyfile,
    creation_date,
    get_file_hash,
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    mkdirp,
    print_json,
    read_file,
    read_json,
    read_yaml,
    recursive_find,
    remove_to_base,
    workdir,
    write_file,
    write_json,
    write_yaml,
)
from .misc import choose, chunks, get_hash, mb_to_bytes, print_bytes, slugify
from .terminal import (
    check_install,
    confirm_action,
    confirm_uninstall,
    ensure_no_extra,
    get_installdir,
    run_command,
)



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


import os

import fluxburst.utils as utils

install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# Default flux operator yaml for kubernetes
flux_operator_yaml = "https://raw.githubusercontent.com/flux-framework/flux-operator/main/examples/dist/flux-operator.yaml"

# Default plugin prefix
plugin_prefix = "fluxburst_"

# User home
userhome = os.path.expanduser("~/.fluxburst")

# The default GitHub registry with recipes (for docgen)
github_url = "https://github.com/converged-computing/flux-burst"



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)


# We define a FluxHandle class to also provide a mock handle,
# meaning we aren't running flux, but can provide fake jobs


class FluxMock:
    """
    A mock flux handle that will return fake jobs, etc.

    This class does not require importing Flux and can be used
    for testing cases or similar.
    """

    def __init__(self, *args, **kwargs):
        pass

    def update_jobspec(self, job):
        """
        Update a jobspec via the kvs
        """
        pass

    def list_jobs(self):
        """
        List one fake, burstable job.

        Generated via:
        flux submit -N 4 --cwd /tmp --setattr=burstable hostname
        in the fluxrm/flux-sched:focal container on July 2nd 2023. Note that we
        only use the high level attributes so hosts, etc. do not matter.
        """
        return {
            "jobs": [
                {
                    "id": 17839985524736,
                    "userid": 1002,
                    "urgency": 16,
                    "priority": 16,
                    "t_submit": 1688329701.680035,
                    "t_depend": 1688329701.693167,
                    "t_run": 1688329701.709494,
                    "t_cleanup": 1688329701.7517478,
                    "t_inactive": 1688329701.753637,
                    "state": 64,
                    "name": "hostname",
                    "cwd": "/tmp",
                    "ntasks": 4,
                    "ncores": 16,
                    "duration": 0.0,
                    "nnodes": 4,
                    "ranks": "[0-3]",
                    "nodelist": "84bd2c990b[15,15,15,15]",
                    "success": True,
                    "exception_occurred": False,
                    "result": 1,
                    "expiration": 4841929701.0,
                    "waitstatus": 0,
                }
            ]
        }

    def get_job_info(self, jobid):
        """
        Get job info. This is job info (the same function called on the container)
        job returned above) in the fluxrm/flux-sched:focal container.
        """
        return {
            "id": 17839985524736,
            "userid": 1002,
            "urgency": 16,
            "priority": 16,
            "t_submit": 1688329701.680035,
            "t_depend": 1688329701.693167,
            "t_run": 1688329701.709494,
            "t_cleanup": 1688329701.7517478,
            "t_inactive": 1688329701.753637,
            "state": 64,
            "name": "hostname",
            "cwd": "/tmp",
            "ntasks": 4,
            "ncores": 16,
            "duration": 0.0,
            "nnodes": 4,
            "ranks": "[0-3]",
            "nodelist": "84bd2c990b[15,15,15,15]",
            "success": True,
            "exception_occurred": False,
            "result": 1,
            "expiration": 4841929701.0,
            "waitstatus": 0,
            "info": {
                "_t_depend": 1688329701.693167,
                "_t_run": 1688329701.709494,
                "_t_cleanup": 1688329701.7517478,
                "_t_inactive": 1688329701.753637,
                "_duration": 0.0,
                "_expiration": 4841929701.0,
                "_name": "hostname",
                "_cwd": "/tmp",
                "_queue": "",
                "_ntasks": 4,
                "_ncores": 16,
                "_nnodes": 4,
                "_priority": 16,
                "_ranks": "[0-3]",
                "_nodelist": "84bd2c990b[15,15,15,15]",
                "_success": True,
                "_waitstatus": 0,
                "_id": 17839985524736,
                "_userid": 1002,
                "_urgency": 16,
                "_t_submit": 1688329701.680035,
                "_exception_occurred": False,
                "_state_id": 64,
                "_result_id": 1,
                "_exception": {
                    "occurred": False,
                    "severity": "",
                    "type": "",
                    "note": "",
                },
                "_annotations": {"annotationsDict": {}, "atuple": ()},
                "_sched": None,
                "_user": None,
                "_dependencies": [],
            },
            "spec": {
                "resources": [
                    {
                        "type": "node",
                        "count": 4,
                        "exclusive": True,
                        "with": [
                            {
                                "type": "slot",
                                "count": 1,
                                "with": [{"type": "core", "count": 1}],
                                "label": "task",
                            }
                        ],
                    }
                ],
                "tasks": [
                    {"command": ["hostname"], "slot": "task", "count": {"per_slot": 1}}
                ],
                "attributes": {
                    "system": {
                        "duration": 0,
                        "cwd": "/tmp",
                        "shell": {
                            "options": {
                                "rlimit": {
                                    "cpu": -1,
                                    "fsize": -1,
                                    "data": -1,
                                    "stack": 8388608,
                                    "core": -1,
                                    "nofile": 1048576,
                                    "as": -1,
                                    "rss": -1,
                                    "nproc": -1,
                                }
                            }
                        },
                        "burstable": 1,
                    }
                },
                "version": 1,
            },
        }


class FluxHandle:
    def __init__(self, handle=None):
        import flux

        self.handle = handle or flux.Flux()

    def update_jobspec(self, job):
        """
        Update a jobspec via the kvs
        """
        import flux.job

        # This is a workaround because updating the attribute directly
        # does not stick yet.
        kvs = flux.job.job_kvs(self.handle, job["id"])
        kvs[kvs.key_at("jobspec")] = job["spec"]
        kvs.commit()
        return kvs

    def list_jobs(self):
        """
        List actual jobs from the flux.job module
        """
        import flux.job

        return flux.job.job_list(self.handle).get()

    def get_job_info(self, jobid):
        """
        Get job info based on an id

        Also retrieve the full job info and jobspec.
        This is not yet currently perfectly json serializable, need to
        handle EmptyObject if that is desired.
        """
        import flux.job

        fluxjob = flux.job.JobID(jobid)
        payload = {"id": fluxjob, "attrs": ["all"]}
        rpc = flux.job.list.JobListIdRPC(self.handle, "job-list.list-id", payload)
        job = rpc.get_job()

        # Job info, timing, priority, etc.
        job["info"] = rpc.get_jobinfo().__dict__
        job["info"]["_exception"] = job["info"]["_exception"].__dict__
        job["info"]["_annotations"] = job["info"]["_annotations"].__dict__

        # the KVS will have annotations!
        kvs = flux.job.job_kvs(self.handle, jobid)
        job["spec"] = kvs.get("jobspec")
        return job



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

__version__ = "0.0.13"
AUTHOR = "Vanessa Sochat"
EMAIL = "vsoch@users.noreply.github.com"
NAME = "flux-burst"
PACKAGE_URL = "https://github.com/converged-computing/flux-burst"
KEYWORDS = "cloud, flux-framework, flux, bursting"
DESCRIPTION = "Flux module that enables burstable plugins."
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (
    ("ruamel.yaml", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    ("rich", {"min_version": None}),
    ("oras", {"min_version": None}),
    ("requests", {"min_version": None}),
    ("pick", {"min_version": None}),
)

INSTALL_REQUIRES_KUBERNETES = (
    ("kubernetes", {"min_version": None}),
    ("fluxoperator", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + INSTALL_REQUIRES_KUBERNETES + TESTS_REQUIRES
